from batou.utils import flatten
from collections import defaultdict


class Subscription(object):

    # Dirty on the subscription means: I am OK to be dirty and _not_
    # get updated. The default is False: I want to be updated.

    def __init__(self, root, strict, host, reverse, dirty):
        self.root = root
        self.strict = strict
        self.host = host
        self.reverse = reverse
        self.dirty = dirty

    def __hash__(self):
        return hash((self.root, self.strict, self.host, self.reverse,
                    self.dirty))


class Resources(object):
    """A registry for resources.

    Resources are mappings of keys to lists of values.

    Components can `provide` resources by specifying a key and a value.

    Components can retrieve all resources for a key by `requiring` the key.
    Optionally components can limit the values returned to a specific host.

    The registry keeps track of which component depends on what
    resources.
    """

    # Maps keys to root components that depend on the key.

    # A "strict" dependency means that it's not OK to have this dependency
    # unsatisfied and requires at least one value.

    # A reverse dependency means that the dependency map will reverse the root
    # components order: "require before provide" instead of the default
    # "provide before require"

    # {key: [(root, strict, host, reverse),
    #        (root, strict, host, reverse), ...]}
    subscribers = None
    # Keeps track of root components that have not seen changes to a key they
    # have subscribed to when they were configured earlier..
    dirty_dependencies = None

    # {key: {host: [values]}}
    resources = None

    def __init__(self):
        self.resources = {}
        self.subscribers = {}
        self.dirty_dependencies = set()

    def _subscriptions(self, key, host):
        return [s for s in self.subscribers.get(key, ())
                if s.host is None or host is None or s.host is host]

    @property
    def strict_subscribers(self):
        for key, subscribers in self.subscribers.items():
            if any(s.strict for s in subscribers):
                yield key

    def provide(self, root, key, value):
        values = self.resources.setdefault(key, defaultdict(list))
        values[root].append(value)
        self.dirty_dependencies.update(
            [s.root for s in self._subscriptions(key, root.host)
             if not s.dirty])

    def get(self, key, host=None):
        """Return resource values without recording a dependency."""
        if host is not None:
            results = []
            for root, values in self.resources.get(key, {}).items():
                if root.component.host is host:
                    results.extend(values)
        else:
            results = flatten(self.resources.get(key, {}).values())
        return results

    def require(self, root, key, host=None, strict=True, reverse=False,
                dirty=False):
        """Return resource values and record component dependency."""
        s = Subscription(root, strict, host, reverse, dirty)
        self.subscribers.setdefault(key, set()).add(s)
        return self.get(key, host)

    def reset_component_resources(self, root):
        """Move all resources aside that were provided by this component."""
        for key, resources in self.resources.items():
            if root not in resources:
                continue
            del resources[root]
            # Removing this resource requires invalidating components that
            # depend on this resource and have already been configured so we
            # need to mark them as dirty if they want to be clean.
            s = [s.root for s in self._subscriptions(key, root.host)
                 if not s.dirty]
            self.dirty_dependencies.update(s)

    def copy_resources(self):
        # A "one level deep" copy of the resources dict to be used by the
        # `unused` property.
        resources = {}
        for key, providers in self.resources.items():
            resources[key] = dict(providers)
        return resources

    @property
    def unused(self):
        # XXX. Gah. This makes my head explode: we need to take the 'host'
        # filter into account whether some of the values provided where never
        # used.
        resources = self.copy_resources()
        for key, subscribers in self.subscribers.items():
            if key not in resources:
                continue
            for s in subscribers:
                if s.host is None:
                    del resources[key]
                    break
                for resource_root in list(resources[key]):
                    if resource_root.host is s.host:
                        del resources[key][resource_root]
                        if not resources[key]:
                            del resources[key]
                            break
                if key not in resources:
                    break
        return resources

    @property
    def unsatisfied(self):
        return set(self.strict_subscribers) - set(self.resources)

    @property
    def unsatisfied_components(self):
        components = set()
        for resource in self.unsatisfied:
            components.update(
                [s.root for s in self._subscriptions(resource, None)])
        return components

    def get_dependency_graph(self):
        """Return a dependency graph as a dict of lists:

        {component: [dependency1, dependency2],
         ...}

        """
        graph = defaultdict(set)
        for key, providers in self.resources.items():
            for s in self._subscriptions(key, None):
                if s.reverse:
                    for provider in providers:
                        graph[provider].add(s.root)
                else:
                    graph[s.root].update(providers)
        return graph



from .component import load_components_from_file
from .host import LocalHost, RemoteHost
from .resources import Resources
from .secrets import add_secrets_to_environment_override
from ConfigParser import RawConfigParser
from batou import MissingComponent
from batou import MissingEnvironment, ComponentLoadingError, SuperfluousSection
from batou import NonConvergingWorkingSet, UnusedResources, ConfigurationError
from batou import SuperfluousComponentSection, CycleErrorDetected
from batou import UnknownComponentConfigurationError, UnsatisfiedResources
from batou import DuplicateHostError
from batou._output import output
from batou.component import RootComponent
from batou.repository import Repository
from batou.utils import cmd
from batou.utils import revert_graph, topological_sort, CycleError
import ast
import batou.c
import batou.vfs
import glob
import os
import os.path
import pwd
import sys


class ConfigSection(dict):
    def as_list(self, option):
        result = self[option]
        if ',' in result:
            result = [x.strip() for x in result.split(',')]
        elif '\n' in result:
            result = (x.strip() for x in result.split('\n'))
            result = [x for x in result if x]
        else:
            result = [result]
        return result


class Config(object):

    def __init__(self, path):
        config = RawConfigParser()
        config.optionxform = lambda s: s
        if path:  # Test support
            config.read(path)
        self.config = config

    def __contains__(self, section):
        return self.config.has_section(section)

    def __getitem__(self, section):
        if section not in self:
            raise KeyError(section)
        return ConfigSection(
            (x, self.config.get(section, x))
            for x in self.config.options(section))

    def __iter__(self):
        return iter(self.config.sections())

    def get(self, section, default=None):
        try:
            return self[section]
        except KeyError:
            return default


class Environment(object):
    """An environment assigns components to hosts and provides
    environment-specific configuration for components.
    """

    service_user = None
    host_domain = None
    branch = None
    connect_method = None
    update_method = None
    platform = None
    vfs_sandbox = None
    timeout = None
    target_directory = None

    repository_url = None
    version = None
    develop = None

    def __init__(self, name, timeout=None, platform=None, basedir='.'):
        self.name = name
        self.hosts = {}
        self.resources = Resources()
        self.overrides = {}
        self.exceptions = []
        self.timeout = timeout
        self.platform = platform

        # These are the component classes, decorated with their
        # name.
        self.components = {}
        # These are the components assigned to hosts.
        self.root_components = []

        self.base_dir = os.path.abspath(basedir)
        self.workdir_base = os.path.join(self.base_dir, 'work')

    def load(self):
        # Scan all components
        for filename in sorted(glob.glob(
                os.path.join(self.base_dir, 'components/*/component.py'))):
            try:
                self.components.update(load_components_from_file(filename))
            except Exception as e:
                self.exceptions.append(ComponentLoadingError(filename, e))

        # Load environment configuration
        config_file = os.path.join(
            self.base_dir, 'environments/{}.cfg'.format(self.name))
        if not os.path.isfile(config_file):
            raise MissingEnvironment(self)

        config = Config(config_file)

        self.load_environment(config)
        self.load_hosts(config)

        # load overrides
        for section in config:
            if section.startswith('host:'):
                continue
            if not section.startswith('component:'):
                if section not in ['hosts', 'environment', 'vfs']:
                    self.exceptions.append(SuperfluousSection(section))
                continue
            root_name = section.replace('component:', '')
            if root_name not in self.components:
                self.exceptions.append(SuperfluousComponentSection(root_name))
                continue
            self.overrides.setdefault(root_name, {})
            self.overrides[root_name].update(config[section])

        self.repository = Repository.from_environment(self)

        # The deployment base is the path relative to the
        # repository where batou is located (with ./batou,
        # ./environments, and ./components)
        self.deployment_base = os.path.relpath(
            self.base_dir, self.repository.root)

    def load_secrets(self):
        add_secrets_to_environment_override(self)

    def load_environment(self, config):
        environment = config.get('environment', {})
        for key in ['service_user', 'host_domain', 'target_directory',
                    'connect_method', 'update_method', 'branch',
                    'platform', 'timeout', 'develop', 'repository_url']:
            if key not in environment:
                continue
            if getattr(self, key) is not None:
                # Avoid overriding early changes that have already been
                # applied, e.g. by tests.
                continue
            setattr(self, key, environment[key])

        self._set_defaults()

        if 'vfs' in config:
            sandbox = config['vfs']['sandbox']
            sandbox = getattr(batou.vfs, sandbox)(self, config['vfs'])
            self.vfs_sandbox = sandbox

    def load_hosts(self, config):
        self._load_hosts_single_section(config)
        self._load_hosts_multi_secion(config)

    def _load_hosts_single_section(self, config):
        for literal_hostname in config.get('hosts', {}):
            hostname = literal_hostname.strip('!')
            host = self.add_host(hostname)
            host.ignore = literal_hostname.startswith('!')
            host.platform = self.platform
            self._load_host_components(
                hostname,
                config['hosts'].as_list(literal_hostname))

    def _load_hosts_multi_secion(self, config):
        for section in config:
            if not section.startswith('host:'):
                continue
            hostname = section.replace('host:', '', 1)
            if hostname in self.hosts:
                self.exceptions.append(DuplicateHostError(hostname))
            host = self.add_host(hostname)
            host.ignore = ast.literal_eval(
                config[section].get('ignore', 'False'))
            host.platform = config[section].get('platform', self.platform)
            self._load_host_components(
                hostname,
                config[section].as_list('components'))

    def _load_host_components(self, hostname, component_list):
            components = parse_host_components(component_list)
            for component, settings in components.items():
                try:
                    self.add_root(component, hostname,
                                  settings['features'], settings['ignore'])
                except KeyError:
                    self.exceptions.append(
                        MissingComponent(component, hostname))

    def _set_defaults(self):
        if self.update_method is None:
            self.update_method = 'rsync'
        if self.connect_method is None:
            self.connect_method = 'ssh'

        if self.service_user is None:
            self.service_user = pwd.getpwuid(os.getuid()).pw_name

        if self.target_directory is None:
            self.target_directory = '~/deployment'

        if self.platform is None and self.host_domain:
            self.platform = self.host_domain

        if self.timeout is None:
            self.timeout = 3
        else:
            self.timeout = int(self.timeout)

        if self.version is None:
            self.version = os.environ.get('BATOU_VERSION')

        if self.develop is None:
            self.develop = os.environ.get('BATOU_DEVELOP')

    # API to instrument environment config loading

    def add_host(self, hostname):
        fqdn = self.normalize_host_name(hostname)
        if fqdn not in self.hosts:
            if self.connect_method == 'local':
                self.hosts[fqdn] = LocalHost(fqdn, self)
            else:
                self.hosts[fqdn] = RemoteHost(fqdn, self)
        return self.hosts[fqdn]

    def add_root(self, component_name, hostname, features=(), ignore=False):
        host = self.add_host(hostname)
        compdef = self.components[component_name]
        root = RootComponent(
            name=compdef.name,
            environment=self,
            host=host,
            features=features,
            ignore=ignore,
            factory=compdef.factory,
            defdir=compdef.defdir,
            workdir=os.path.join(self.workdir_base, compdef.name))
        self.root_components.append(root)
        return root

    def get_root(self, component_name, hostname):
        host = self.add_host(hostname)
        for root in self.root_components:
            if root.host == host and root.name == component_name:
                return root
        raise KeyError("Component {} not configured for host {}".format(
            component_name, hostname))

    def prepare_connect(self):
        if self.deployment.fast:
            return
        if self.connect_method == 'vagrant':
            output.step("vagrant", "Ensuring machines are up ...")
            cmd('vagrant up')

    # Deployment API (implements the configure-verify-update cycle)

    def configure(self):
        """Configure all root components.

        Monitor the dependencies between resources and try to reach a stable
        order.

        """
        working_set = set(self.root_components)

        previous_working_sets = []
        exceptions = []

        while working_set:
            exceptions = []
            previous_working_sets.append(working_set.copy())
            retry = set()
            self.resources.dirty_dependencies.clear()

            for root in working_set:
                try:
                    self.resources.reset_component_resources(root)
                    root.overrides = self.overrides.get(root.name, {})
                    root.prepare()
                except ConfigurationError as e:
                    # A known exception which we can report gracefully later.
                    exceptions.append(e)
                    retry.add(root)
                except Exception as e:
                    # An unknown exception which we have to work harder
                    # to report gracefully.
                    ex_type, ex, tb = sys.exc_info()
                    exceptions.append(
                        UnknownComponentConfigurationError(root, e, tb))
                    retry.add(root)

            retry.update(self.resources.dirty_dependencies)
            retry.update(self.resources.unsatisfied_components)

            # Try to find a valid order of the components. If we can't then we
            # have detected a dependency cycle and need to stop.

            try:
                self.roots_in_order()
            except CycleError as e:
                exceptions.append(CycleErrorDetected(e))

            if (retry in previous_working_sets):
                # If any resources were required, now is the time to report
                # them.
                if self.resources.unsatisfied:
                    exceptions.append(UnsatisfiedResources(
                        self.resources.unsatisfied))

                # We did not manage to improve on our last working set, so we
                # give up.
                exceptions.append(NonConvergingWorkingSet(retry))
                break

            working_set = retry

        # We managed to converge on a working set. However, some resource were
        # provided but never used. We're rather picky here and report this as
        # an error.
        if self.resources.unused:
            exceptions.append(UnusedResources(self.resources.unused))

        self.exceptions.extend(exceptions)
        if self.exceptions:
            # We just raise here to support a reasonable flow
            # for our caller. We expect him to look at our exceptions
            # attribute anyway.
            raise self.exceptions[0]

    def roots_in_order(self, host=None):
        """Return a list of root components sorted by their resource
        dependencies, filtered by host if specified.

        Raises ValueError if component dependencies have cycles.

        """
        dependencies = self.resources.get_dependency_graph()
        # Complete the graph with components that do not have any dependencies
        # (yet)
        for root in self.root_components:
            if root not in dependencies:
                dependencies[root] = set()
        roots = topological_sort(revert_graph(dependencies))
        if host is not None:
            host = self.get_host(host)
            roots = filter(lambda x: x.host is host, roots)
        return roots

    def normalize_host_name(self, hostname):
        """Ensure the given host name is an FQDN for this environment."""
        if not self.host_domain or hostname.endswith(self.host_domain):
            return hostname
        else:
            return '%s.%s' % (hostname, self.host_domain)

    def get_host(self, hostname):
        return self.hosts[self.normalize_host_name(hostname)]

    def map(self, path):
        if self.vfs_sandbox:
            return self.vfs_sandbox.map(path)
        return path


def parse_host_components(components):
    """Parse a component list as given in an environment config for a host
    into a dict of dicts:

        {'name': {'features': [], 'ignore': False}}

    If one component is ignored, then the whole set of component features
    is ignored.

    Expected syntax:

    [!]component[:feature], component[:feature]

    """
    result = {}
    for name in components:
        name = name.strip()
        if ':' in name:
            name, feature = name.split(':', 1)
        else:
            feature = None

        ignore = name.startswith('!')
        name = name.strip('!')
        result.setdefault(name, {'features': [], 'ignore': False})
        result[name]['ignore'] |= ignore
        if feature:
            result[name]['features'].append(feature)
    return result



from StringIO import StringIO
from batou import MissingEnvironment
from batou.environment import Environment, Config
from mock import Mock
import mock
import pytest


def test_environment_should_raise_if_no_config_file(tmpdir):
    e = Environment(u'foobar')
    with pytest.raises(MissingEnvironment):
        e.load()


def test_load_should_use_defaults(sample_service):
    e = Environment(u'test-without-env-config')
    e.load()
    assert e.host_domain is None
    assert e.branch is None


def test_load_should_use_config(sample_service):
    e = Environment(u'test-with-env-config')
    e.load()
    assert e.service_user == u'joe'
    assert e.host_domain == u'example.com'
    assert e.branch == u'release'


def test_load_ignores_predefined_environment_settings(sample_service):
    e = Environment(u'test-with-env-config')
    e.service_user = u'bob'
    e.host_domain = u'sample.com'
    e.branch = u'default'
    e.load()
    assert e.service_user == u'bob'
    assert e.host_domain == u'sample.com'
    assert e.branch == u'default'


def test_load_sets_up_overrides(sample_service):
    e = Environment('test-with-overrides')
    e.load()
    assert e.overrides == {'hello1': {'asdf': '1'}}


def test_loads_configures_vfs_sandbox(sample_service):
    e = Environment('test-with-vfs-sandbox')
    e.load()
    assert e.vfs_sandbox.map('/asdf').endswith('work/_/asdf')
    assert e.map('/asdf').endswith('work/_/asdf')


def test_default_sandbox_is_identity():
    e = Environment(u'foo')
    assert e.map('/asdf') == '/asdf'


def test_get_host_raises_keyerror_if_unknown():
    e = Environment(u'name')
    with pytest.raises(KeyError):
        e.get_host('asdf')


def test_get_host_normalizes_hostname():
    e = Environment(u'name')
    e.hosts['asdf.example.com'] = host = Mock()
    e.host_domain = 'example.com'
    assert host == e.get_host('asdf')
    assert host == e.get_host('asdf.example.com')


def test_normalize_hostname_regression_11156():
    # The issue here was that we used "rstrip" which works on characters,
    # not substrings. Having the domain (example.com) start with an eee
    # causes the hostname to get stripped of it's "eee"s ending in an
    # empty hostname accidentally.
    e = Environment(u'name')
    e.hosts['eee.example.com'] = host = Mock()
    e.host_domain = 'example.com'
    assert host == e.get_host('eee')


def test_get_host_without_subdomain_also_works():
    e = Environment(u'name')
    e.hosts['example.com'] = host = Mock()
    e.host_domain = 'example.com'
    assert host == e.get_host('example.com')


def test_get_root_raises_keyerror_on_nonassigned_component():
    e = Environment(u'foo')
    with pytest.raises(KeyError):
        e.get_root('asdf', 'localhost')


def test_multiple_components(sample_service):
    e = Environment(u'test-multiple-components')
    e.load()
    components = dict(
        (host, list(sorted(c.name for c in e.roots_in_order(host=host))))
        for host in sorted(e.hosts.keys()))
    assert components == dict(
        localhost=['hello1', 'hello2'],
        otherhost=['hello3', 'hello4'],
        thishost=['hello5', 'hello6'])


def test_parse_host_components():
    from batou.environment import parse_host_components
    assert (parse_host_components(['asdf']) ==
            {'asdf': {'features': [], 'ignore': False}})

    assert (parse_host_components(['!asdf']) ==
            {'asdf': {'features': [], 'ignore': True}})

    assert (parse_host_components(['!asdf:test', 'asdf:bar']) ==
            {'asdf': {'features': ['test', 'bar'], 'ignore': True}})

    assert (parse_host_components(['asdf:test', 'asdf:bar']) ==
            {'asdf': {'features': ['test', 'bar'], 'ignore': False}})


@mock.patch('batou.environment.Environment.add_root')
def test_load_hosts_should_merge_single_and_multi_definition(add_root):
    e = Environment(u'name')
    config = Config(None)
    config.config.readfp(StringIO("""
[hosts]
foo = bar
[host:baz]
components = bar
    """))
    e.load_hosts(config)
    assert [mock.call('bar', 'foo', [], False),
            mock.call('bar', 'baz', [], False)] == \
        add_root.call_args_list


@mock.patch('batou.environment.Environment.add_root')
def test_load_hosts_should_load_single_hosts_section(add_root):
    e = Environment(u'name')
    config = Config(None)
    config.config.readfp(StringIO("""
[hosts]
foo = bar
    """))
    e.load_hosts(config)
    add_root.assert_called_once_with('bar', 'foo', [], False)


@mock.patch('batou.environment.Environment.add_root')
def test_load_hosts_should_load_multi_hosts_section(add_root):
    pass
    e = Environment(u'name')
    config = Config(None)
    config.config.readfp(StringIO("""
[host:foo]
components = bar
    """))
    e.load_hosts(config)
    add_root.assert_called_once_with('bar', 'foo', [], False)


@mock.patch('batou.environment.Environment.add_root')
def test_load_hosts_multi_should_use_ignore_flag(add_root):
    pass
    e = Environment(u'name')
    config = Config(None)
    config.config.readfp(StringIO("""
[host:foo]
components = bar
ignore = True
    """))
    e.load_hosts(config)
    assert e.hosts['foo'].ignore


@mock.patch('batou.environment.Environment.add_root')
def test_load_hosts_should_break_on_duplicate_definition(add_root):
    e = Environment(u'name')
    config = Config(None)
    config.config.readfp(StringIO("""
[hosts]
foo = bar
[host:foo]
components = bar
    """))
    e.load_hosts(config)
    assert e.exceptions
    assert 'foo' == e.exceptions[0].hostname


@mock.patch('batou.environment.Environment.add_root')
def test_load_hosts_multi_should_use_env_platform(add_root):
    pass
    e = Environment(u'name')
    e.platform = mock.sentinel.platform
    config = Config(None)
    config.config.readfp(StringIO("""
[host:foo]
components = bar
ignore = True
    """))
    e.load_hosts(config)
    assert e.hosts['foo'].platform == mock.sentinel.platform


@mock.patch('batou.environment.Environment.add_root')
def test_load_hosts_multi_should_use_host_platform_if_given(add_root):
    pass
    e = Environment(u'name')
    e.platform = mock.sentinel.platform
    config = Config(None)
    config.config.readfp(StringIO("""
[host:foo]
components = bar
platform = specific
    """))
    e.load_hosts(config)
    assert e.hosts['foo'].platform == 'specific'


@mock.patch('batou.environment.Environment.add_root')
def test_load_hosts_single_should_use_env_platform(add_root):
    e = Environment(u'name')
    e.platform = mock.sentinel.platform
    config = Config(None)
    config.config.readfp(StringIO("""
[hosts]
foo = bar
    """))
    e.load_hosts(config)
    assert e.hosts['foo'].platform == mock.sentinel.platform



import os
import os.path
import pwd
import subprocess
import traceback


# Satisfy flake8 and support testing.
try:
    channel
except NameError:
    channel = None


deployment = None
environment = None
target_directory = None


# The output class should really live in _output. However, to support
# bootstrapping we define it here and then re-import in the _output module.


class Output(object):
    """Manage the output of various parts of batou to achieve
    consistency wrt to formatting and display.
    """

    enable_debug = False

    def __init__(self, backend):
        self.backend = backend

    def line(self, message, debug=False, **format):
        if debug and not self.enable_debug:
            return
        self.backend.line(message, **format)

    def annotate(self, message, debug=False, **format):
        if debug and not self.enable_debug:
            return
        lines = message.split('\n')
        lines = [' ' * 5 + line for line in lines]
        message = '\n'.join(lines)
        self.line(message, **format)

    def tabular(self, key, value, separator=': ', debug=False, **kw):
        if debug and not self.enable_debug:
            return
        message = key.rjust(10) + separator + value
        self.annotate(message, **kw)

    def section(self, title, debug=False, **format):
        if debug and not self.enable_debug:
            return
        _format = {'bold': True}
        _format.update(format)
        self.backend.sep("=", title, **_format)

    def sep(self, sep, title, **format):
        return self.backend.sep(sep, title, **format)

    def step(self, context, message, debug=False, **format):
        if debug and not self.enable_debug:
            return
        _format = {'bold': True}
        _format.update(format)
        self.line('{}: {}'.format(context, message), **_format)

    def error(self, message, exc_info=None, debug=False):
        if debug and not self.enable_debug:
            return
        self.step("ERROR", message, red=True)
        if exc_info:
            tb = traceback.format_exception(*exc_info)
            tb = ''.join(tb)
            tb = '      ' + tb.replace('\n', '\n      ') + '\n'
            self.backend.write(tb, red=True)


class ChannelBackend(object):

    def __init__(self, channel):
        self.channel = channel

    def _send(self, output_cmd, *args, **kw):
        self.channel.send(('batou-output', output_cmd, args, kw))

    def line(self, message, **format):
        self._send('line', message, **format)

    def sep(self, sep, title, **format):
        self._send('sep', sep, title, **format)

    def write(self, content, **format):
        self._send('write', content, **format)


class Deployment(object):

    environment = None

    def __init__(self, env_name, host_name, overrides,
                 timeout, platform):
        self.env_name = env_name
        self.host_name = host_name
        self.overrides = overrides
        self.timeout = timeout
        self.platform = platform

    def load(self):
        from batou.environment import Environment
        self.environment = Environment(
            self.env_name, self.timeout, self.platform)
        self.environment.deployment = self
        self.environment.load()
        self.environment.overrides = self.overrides
        self.environment.configure()

    def deploy(self, root, predict_only):
        root = self.environment.get_root(root, self.host_name)
        root.component.deploy(predict_only)


def lock():
    # XXX implement!
    pass


class CmdError(Exception):

    def __init__(self, cmd, returncode, stdout, stderr):
        self.cmd = cmd
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr

    def report(self):
        output.error(self.cmd)
        output.tabular("Return code", str(self.returncode), red=True)
        output.line('STDOUT', red=True)
        output.annotate(self.stdout)
        output.line('STDERR', red=True)
        output.annotate(self.stderr)


def cmd(c, acceptable_returncodes=[0]):
    process = subprocess.Popen(
        ['LANG=C LC_ALL=C LANGUAGE=C {}'.format(c)],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE,
        shell=True)
    stdout, stderr = process.communicate()
    if process.returncode not in acceptable_returncodes:
        raise CmdError(c, process.returncode, stdout, stderr)
    return stdout, stderr


def ensure_repository(target, method):
    target = os.path.expanduser(target)
    global target_directory
    target_directory = target

    if not os.path.exists(target):
        os.makedirs(target)

    if method in ['hg-pull', 'hg-bundle']:
        if not os.path.exists(target + '/.hg'):
            cmd("hg init {}".format(target))
    elif method in ['git-pull', 'git-bundle']:
        if not os.path.exists(target + '/.git'):
            cmd("git init {}".format(target))
    elif method == 'rsync':
        pass
    else:
        raise RuntimeError("Unknown repository method: {}".format(method))

    return target


def ensure_base(base):
    base = os.path.join(target_directory, base)
    if not os.path.exists(base):
        os.makedirs(base)
    return base


def hg_current_heads():
    target = target_directory
    os.chdir(target)
    result = []
    id, _ = cmd('hg id -i')
    id = id.strip()
    if id == '000000000000':
        return [id]
    heads, _ = cmd('hg heads')
    for line in heads.split('\n'):
        if not line.startswith('changeset:'):
            continue
        line = line.split(':')
        result.append(line[2])
    return result


def hg_pull_code(upstream):
    # TODO Make choice of VCS flexible
    target = target_directory
    os.chdir(target)
    # Phase 1: update working copy
    # XXX manage certificates
    cmd("hg pull {}".format(upstream))


def hg_unbundle_code():
    # TODO Make choice of VCS flexible
    # XXX does this protect us from accidental new heads?
    target = target_directory
    os.chdir(target)
    cmd('hg -y unbundle batou-bundle.hg')


def hg_update_working_copy(branch):
    cmd("hg up -C {}".format(branch))
    id, _ = cmd("hg id -i")
    id = id.strip()
    return id


def git_current_head(branch):
    target = target_directory
    os.chdir(target)
    id, err = cmd('git rev-parse {branch}'.format(branch=branch),
                  acceptable_returncodes=[0, 128])
    id = id.strip()
    return id if 'unknown revision' not in err else None


def git_pull_code(upstream, branch):
    target = target_directory
    os.chdir(target)
    out, err = cmd('git remote -v')
    for line in out.splitlines():
        line = line.strip()
        line = line.replace(' ', '\t')
        if not line:
            continue
        name, remote, _ = line.split('\t', 3)
        if name != 'batou-pull':
            continue
        if remote != upstream:
            cmd('git remote remove batou-pull')
        # The batou-pull remote is correctly configured.
        break
    else:
        cmd('git remote add batou-pull {upstream}'.format(upstream=upstream))
    cmd('git fetch batou-pull')


def git_unbundle_code():
    target = target_directory
    os.chdir(target)
    out, err = cmd('git remote -v')
    if 'batou-bundle' not in out:
        cmd('git remote add batou-bundle batou-bundle.git')
    cmd('git fetch batou-bundle')


def git_update_working_copy(branch):
    cmd('git checkout --force {branch}'.format(branch=branch))
    cmd('git clean -fd')  # merge fails, if there are files around.
    # For git <2.0:
    cmd('git config merge.defaultToUpstream true')
    cmd('git merge --ff-only')
    id, _ = cmd('git rev-parse HEAD')
    return id.strip()


def build_batou(deployment_base, bootstrap, batou_args=()):
    target = target_directory
    os.chdir(os.path.join(target, deployment_base))
    with open('batou', 'w') as f:
        f.write(bootstrap)
    os.chmod('batou', 0o755)

    args = ['--help']
    if batou_args:
        args.extend(args)

    cmd('./batou {}'.format(' '.join(args)))


def setup_deployment(
        deployment_base, *args):
    target = target_directory
    os.chdir(os.path.join(target, deployment_base))
    global deployment
    deployment = Deployment(*args)
    deployment.load()


def deploy(root, predict_only=False):
    deployment.deploy(root, predict_only)


def roots_in_order():
    result = []
    for root in deployment.environment.roots_in_order():
        result.append((root.host.fqdn, root.name, root.ignore))
    return result


def whoami():
    return pwd.getpwuid(os.getuid()).pw_name


def setup_output():
    from batou._output import output
    output.backend = ChannelBackend(channel)


if __name__ == '__channelexec__':
    output = Output(ChannelBackend(channel))
    while not channel.isclosed():
        task, args, kw = channel.receive()
        # Support slow bootstrapping
        try:
            import batou
        except ImportError:
            batou = None
        try:
            result = locals()[task](*args, **kw)
            channel.send(('batou-result', result))
        except getattr(batou, 'ConfigurationError', None) as e:
            if e not in deployment.environment.exceptions:
                deployment.environment.exceptions.append(e)
            # Report on why configuration failed.
            deployment.environment.exceptions.sort(key=lambda x: x.sort_key)

            for exception in deployment.environment.exceptions:
                if isinstance(e, batou.SilentConfigurationError):
                    continue
                exception.report()

            batou.output.section(
                "{} ERRORS - CONFIGURATION FAILED".format(
                    len(deployment.environment.exceptions)), red=True)
            channel.send(('batou-configuration-error', None))
        except getattr(batou, 'DeploymentError', None) as e:
            e.report()
            channel.send(('batou-deployment-error', None))
        except Exception as e:
            # I voted for duck-typing here as we may be running in the
            # bootstrapping phase and don't have access to all classes yet.
            if hasattr(e, 'report'):
                e.report()
                channel.send(('batou-error', None))
            else:
                tb = traceback.format_exc()
                channel.send(('batou-unknown-error', tb))



from batou.component import Component
from batou.lib.python import VirtualEnv
from batou.update import update_bootstrap
from batou.utils import cmd
import os
import os.path
import pytest


@pytest.mark.slow
@pytest.mark.jenkinsonly
@pytest.mark.timeout(240)
def test_package_venv_installations(root):
    # this is a really nasty test ...
    base_dir = os.path.join(
        os.path.dirname(__file__), '../../../../')

    # copy the "venv" example
    cmd('cp -a {}/examples/venvs/* .'.format(base_dir))
    cmd('rm -rf work')

    # ensure the dev link is OK
    update_bootstrap('', base_dir)

    # run batou, hope for the best. ;)
    stdout, stderr = cmd('./batou deploy dev')
    assert "Deploying component py26" in stdout
    assert "Deploying component py27" in stdout
    assert "Deploying component py33" in stdout
    assert "Deploying component py34" in stdout
    assert "Deploying component py35" in stdout
    assert stderr == ""


def test_venv_updates_if_python_changes(root):
    import ast

    class Playground(Component):
        namevar = 'version'

        def configure(self):
            self.venv = VirtualEnv(self.version)
            self += self.venv

    playground = Playground('2.6')
    root.component += playground
    playground.deploy()
    root.component.sub_components.remove(playground)

    playground = Playground('2.7')
    root.component += playground
    playground.deploy()

    out, err = playground.cmd(
        '{}/bin/python -c "import sys; print sys.version_info[:2]"'.format(
            playground.workdir))
    assert (2, 7) == ast.literal_eval(out)


def test_venv_does_not_update_if_python_does_not_change(root):

    class Playground(Component):
        namevar = 'version'

        def configure(self):
            self.venv = VirtualEnv(self.version)
            self += self.venv

    playground = Playground('2.7')
    root.component += playground
    playground.deploy()
    assert playground.changed
    playground.deploy()
    assert not playground.changed



from batou.vfs import Developer, Map
from mock import Mock
import os.path


def test_developer_mapping(tmpdir):
    environment = Mock()
    environment.workdir_base = str(tmpdir)
    config = Mock()
    mapping = Developer(environment, config)

    assert mapping.map('/etc') == str(tmpdir / '_' / 'etc')
    assert os.path.exists(str(tmpdir / '_'))
    assert not os.path.exists(str(tmpdir / '_' / 'etc'))

    assert mapping.map('/var') == str(tmpdir / '_' / 'var')
    assert not os.path.exists(str(tmpdir / '_' / 'var'))

    assert mapping.map(str(tmpdir / 'foo')) == str(tmpdir / 'foo')
    assert mapping.map('asdf') == 'asdf'


def test_arbitrary_mapping():
    environment = Mock()
    environment.workdir_base = 'workdir'
    config = {'/a': '/a1',
              # longer paths are preferred matches
              '/a/b': '/a2',
              # non-absolute paths are ignored
              'etc': 'none'}
    mapping = Map(environment, config)

    assert mapping.map('/etc') == '/etc'
    assert mapping.map('/a') == '/a1'
    assert mapping.map('/a/b/c') == '/a2/c'
    assert mapping.map('/a/c') == '/a1/c'
    assert mapping.map('workdir/foo') == 'workdir/foo'
